from typing import Optional
from datetime import datetime

from ..utils import (
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
)

from ..ref import (
    AlarmRef,
    CardholderRef,
    DoorRef,
    AccessZoneRef,
    DivisionRef,
    ItemRef,
)


class EventDetail(
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
):
    """Details of an event that took place on a server"""

    server_display_name: str
    time: datetime
    message: Optional[str]
    occurrences: int
    priority: int
    alarm: AlarmRef

    operator: CardholderRef
    source: ItemRef
    # group: EventGroupSummary
    # type: EventTypeSummary
    # event_type: EventTypeSummary
    # division: DivisionRef
    # cardholder: CardholderSummary
    entry_access_zone: AccessZoneRef
    exit_access_zone: AccessZoneRef
    door: DoorRef
    access_group: HrefMixin
    # card: str
    # modified_item: str



""" Alarm Details

The following are used to parse the details of an alarm

"""

from typing import Optional
from datetime import datetime

from ..utils import (
    AppBaseModel,
    HrefMixin,
    OptionalHrefMixin,
    IdentityMixin,
)

from ..ref import (
    CardholderRef,
    OperatorRef,
    InstructionRef,
)

from ..summary import (
    EventTypeSummary,
    AlarmSourceSummary,
)


class AlarmHistoryDetail(
    AppBaseModel,
):
    """Alarm History presented in the details"""

    action: str
    time: datetime
    comment: str  # Comment provided as part of an action
    operator: OperatorRef


class AlarmDetail(
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
):
    """AlarmSummary gives us unactioned events from the CC

    While the detail and summary would typically differ, it seems that
    the API endpoints return more detail in the summary endpoints.

    Note: that we have a number of hrefs that we should follow to
    HATEOAS compliance as per the documentation.

    """

    time: datetime
    message: str
    source: AlarmSourceSummary
    type: str
    event_type: Optional[EventTypeSummary] = None
    priority: int
    state: str
    active: bool
    division: HrefMixin
    event: OptionalHrefMixin = None
    note_presets: list[str] = []

    # The following URLS should be used to follow through
    # on various actions that the system allows
    view: HrefMixin
    comment: HrefMixin
    acknowledge: OptionalHrefMixin = None
    acknowledge_with_comment: OptionalHrefMixin = None
    process: OptionalHrefMixin = None
    process_with_comment: OptionalHrefMixin = None
    force_process: OptionalHrefMixin = None

    # Details fields, these are what don't appear in the summary
    details: Optional[str] = ""
    history: Optional[list[AlarmHistoryDetail]] = []
    instruction: Optional[InstructionRef] = None
    cardholder: Optional[CardholderRef] = None

    def __rich_repr__(self):
        return (
            f"[blue bold] alarm",
            f"{'id':>20} {self.id}",
            f"{'priority':>20} {self.priority}",
            f"{'type':>20} {self.type}",
            f"{'message':>20} {self.message}",
            f"{'active':>20} {'yes' if self.active else 'no'}",
            f"",
            f"[blue bold] hrefs",
            f"{'acknowledge':>20} [link={self.acknowledge.href}]edit[/link]",
        )



""" Personal Data Field Summary
"""
from ..enum import PdfType

from ..utils import (
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
)

class PdfSummary(
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
):
    """ Personal Data Field Summary

    Primarily used as the definition in the Cardholder response
    """
    
    name: str
    type: PdfType # See gallagher docs for enum


""" Adapter and SQLAlchemy dialect for shillelagh

The package provides extensions to provide a SQL interface for the
Gallagher API.

[shillelagh](https://github.com/betodealmeida/shillelagh)
"""
import os
import asyncio
import logging
import urllib
from datetime import datetime

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')
# logging.basicConfig(level=logging.ERROR, format='%(asctime)s %(levelname)s %(message)s')

from typing import (
    Any,
    Dict,
    Iterator,
    AsyncIterator,
    List, 
    Optional,
    Tuple,
)

from shillelagh.adapters.base import Adapter
from shillelagh.typing import (
    RequestedOrder, 
    Row,
)

from shillelagh.exceptions import (
    ImpossibleFilterError,
    InterfaceError,
    InternalError,
    ProgrammingError,
    UnauthenticatedError,
)

from shillelagh.filters import (
    Filter,
    Impossible,
    Operator,
    Range,
)

from shillelagh.fields import (
    Field,
    Float,
    Integer,
    String,
    Order,
    Boolean,
    DateTime,
    Collection,
    Blob,
)

# TODO: refactor this to generic based on SQL.md
from gallagher import cc
from gallagher.cc.alarms import __shillelagh__ \
    as alarms_tables
from gallagher.cc.cardholders import __shillelagh__ \
    as cardholders_tables
from gallagher.cc.status_overrides import __shillelagh__ \
    as status_overrides_tables

class CCAPIAdapter(Adapter):

    # Use this to log messages to assist with shillelagh debugging
    _logger = logging.getLogger(__name__)

    # Concatenate all the tables into a single list for efficiency
    # This is declared here because it's accessed by a static method
    # because shillelagh's supports method is static
    _all_tables = alarms_tables + cardholders_tables + \
        status_overrides_tables

    # Maps the python types to the shillelagh fields
    # initialised here so we don't have to keep redefining it
    _type_map = {
        int: Integer,
        str: String,
        bool: Boolean,
        bytes: Blob,
        list: Collection,
        datetime: DateTime,
        float: Float,
    }

    # API endpoint being used to parse this URL
    # TODO: see if this changes with every instantiation
    _api_endpoint = None

    # The adapter doesn't access the filesystem.
    safe = True

    # The adapter will receive a ``limit`` argument in the ``get_data``
    # method, and will be responsible for limiting the number of rows returned.
    supports_limit = True
    supports_offset = True

    # Check to see if we can do this using the partial column feature
    supports_requested_columns = False

    @staticmethod
    def get_endpoint_urls() -> list[str]:
        # TODO: we need to improve performance by cache this
        # NOTE: this is a method because we need bootstrap to run first
        return [
            f"{table.__config__.endpoint.href}" for table \
                in CCAPIAdapter._all_tables
        ]


    @staticmethod
    def bootstrap_api_client():
        """ Bootstrap the API client for the adapter 
        
        This checks for the availability of the GACC_API_KEY environment
        and sets the variable in the API client
        
        Call this before any operations on the API client
        """
        if not 'GACC_API_KEY' in os.environ:
            raise ValueError(
                "GACC_API_KEY environment variable must be set"
            )

        api_key = os.environ.get('GACC_API_KEY')

        if not api_key:
            raise UnauthenticatedError(
                "GACC_API_KEY environment variable must be set"
            )
        
        cc.api_key = api_key

        for table in CCAPIAdapter._all_tables:
            # This should get us the url, and if not then we are in an invalid state
            # Discover should only ever run for the first endpoint and all
            # others should then deffer to the cached property
            # 
            # Running this here saves us from having to run it in other places
            asyncio.run(table._discover())


    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        """ Return the URL if it is a Gallagher Command Centre API URL

        GACC requires us to respect the HATEOS principle, so we need to
        run discovery before we can evaluate the URL, this means that we
        require the API key to be present and initialise the client.

        Because we share the cc object across multiple methods, it available
        outside the scope of this method
        """
        CCAPIAdapter.bootstrap_api_client()

        if not uri in CCAPIAdapter.get_endpoint_urls():
            CCAPIAdapter._logger.debug(
                f"{uri} not found in {CCAPIAdapter.get_endpoint_urls()}"
            )
            return False

        # Parse the base url using urlparse for comparison        
        base_parsed_url = urllib.parse.urlparse(cc.api_base)

        # Parse the endpoint using urlparse
        parsed_url = urllib.parse.urlparse(uri)

        # Match the netloc property to be equal to the api_base
        # and the scheme to be equal to the base_parsed_url
        #
        # Note that do not match the path as it wil be different
        # and it should have passed the check of being in the endpoint_urls
        if not parsed_url.netloc == base_parsed_url.netloc or\
            not parsed_url.scheme == base_parsed_url.scheme:
            return False
        
        return True
        
    @staticmethod
    def parse_uri(uri: str) -> Tuple[str, str]:
        """ 

        Returns:
            Tuple[str, str]: The URI and the API key
        """
        return (uri, os.environ.get('GACC_API_KEY'))

    def __init__(self, uri: str, api_key: Optional[str], **kwargs: Any):

        super().__init__()

        self.uri = uri
        # TODO: might be redundant due to moving this up the package level
        cc.api_key = api_key

        self._logger.debug(f"Finding suitable adapter for {self.uri}")

        self._api_endpoint = next(
            (table for table in self._all_tables if self.uri == f"{table.__config__.endpoint.href}"),
            None
        )

        if self._api_endpoint:
            self._logger.debug(f"Found helper class = {self._api_endpoint}")
        else:
            self._logger.debug("No suitable adapter found.")

    def get_columns(self) -> Dict[str, Field]:
        """Return a pyndatic DTO as shil compatible attribute config

        Rules here are that we translate as many dictionary vars into
        a __shillelagh__ compatible format.

        If they are hrefs to other children then we select the id field for
        each one of those objects
        """

        if not self._api_endpoint:
            self._logger.debug("No suitable adapter found while get_columns.")
            return {}

        return {
            key: self._type_map[value]() # shillelagh requires an instance
            for key, value in \
                self._api_endpoint.__config__.dto_retrieve._accumulated_annotations()
            if not key.startswith("_") and value in self._type_map
        }


    def get_data(  # pylint: disable=too-many-locals
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        
        dto_list = asyncio.run(self._api_endpoint.list())

        for row in dto_list.result_set:

            yield {
                'rowid': row.id, # Append this for shillelagh
                **row.dict(), # Rest of the responses 
            }




from typing import Optional

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
    OptionalHrefMixin,
)
from ..summary import (
    AlarmSummary,
    AccessSummary,
    SaltoAccessItemSummary,
)
from ..ref import (
    AccessGroupRef,
    PDFRef,
)

from .division import DivisionDetail


class AccessGroupDetail(
    AppBaseModel,
    HrefMixin,
):
    """ """

    name: str
    description: Optional[str]
    parent: Optional[AccessGroupRef]
    division: IdentityMixin
    cardholders: OptionalHrefMixin
    server_display_name: Optional[str]

    description: Optional[str]
    parent: Optional[AccessGroupRef]
    division: DivisionDetail
    cardholders: OptionalHrefMixin
    server_display_name: Optional[str]
    children: list[AccessGroupRef]
    personal_data_definitions: list[PDFRef]

    visitor: bool
    escort_visitors: bool
    lock_unlock_access_zones: bool
    enter_during_lockdown: bool
    first_card_unlock: bool
    override_aperio_privacy: bool
    aperio_offline_access: bool
    disarm_alarm_zones: bool
    arm_alarm_zones: bool
    hvLf_fence_zones: bool
    view_alarms: bool
    shunt: bool
    lock_out_fence_zones: bool
    cancel_fence_zone_lockout: bool
    ack_all: bool
    ack_below_high: bool
    select_alarm_zone: bool
    arm_while_alarm: bool
    arm_while_active_alarm: bool
    isolate_alarm_zones: bool

    access: list[AccessSummary]
    salto_access: list[SaltoAccessItemSummary]
    alarm_zones: list[AlarmSummary]



from typing import Optional

from ..utils import (
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
)

from ..summary import CardExpiryTypeSummary

from ..enum import CredentialsClass


class CardTypeDetail(AppBaseModel, IdentityMixin, HrefMixin):
    """Card Types are cards mobile or physical that are supported at a site"""

    name: str
    minimum_number: Optional[str] = None
    maximum_number: Optional[str] = None
    initial_card_state: str
    facility_code: str
    credential_class: CredentialsClass
    available_card_states: list[str] = None
    default_expiry: Optional[CardExpiryTypeSummary] = None
    send_registration_email: Optional[bool] = False
    send_registration_sms: Optional[bool] = False



"""

"""

from ..utils import (
    AppBaseResponseModel,
)

from ..summary import CardholderSummary


class CardholderSummaryResponse(
    AppBaseResponseModel,
):
    """Summary response for cardholder list and search

    /api/cardholders is generally the endpoint that responds
    to the query, it is dynamically configured from the discovery

    """

    results: list[CardholderSummary]

    @property
    def result_set(self) -> list[CardholderSummary]:
        """ Wrap summary response target property

        the sql interface will call this property and each summary
        response is expected to override this and return the appropriate
        target property
        """
        return self.results

    @property
    def cli_header(self):
        return ("id", "first name", "last name", "authorised")

    def __rich_repr__(self):
        return (r.__rich_repr__() for r in self.results)

    def __str__(self):
        return f"{len(self.results)} cardholders"



from ..utils import (
    AppBaseResponseModel,
)

from ..summary import ScheduleSummary


class ScheduleSummaryResponse(
    AppBaseResponseModel,
):
    """Schedule is a time"""

    results: list[ScheduleSummary]



"""" Gallagher Python idiomatic client

 Copyight (c) 2023, Anomaly Software Pty Ltd

 This project is **NOT** officially affiliated with Gallagher Security.

 Distributed under the terms of the MIT License.
"""

__version__ = "0.1.0-alpha.4"



""" Division

"""

from ..core import Capabilities, APIEndpoint, EndpointConfig

from ...dto.detail import (
    DivisionDetail,
)

from ...dto.response import (
    DivisionSummaryResponse,
)


class Division(APIEndpoint):
    """
    Gallagher advises against hardcoding the URLs for divisions, and instead
    recommends using the /api endpoint to discover the URLs from
    events.divisions.href and alarms.division.href.

    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.divisions.divisions,
            dto_list=DivisionSummaryResponse,
            dto_retrieve=DivisionDetail,
        )



""" Alarms

Command Centre raised alarms for 

"""

from typing import Optional

from ..core import (
    APIEndpoint,
    EndpointConfig,
    Capabilities,
)

from ...dto.ref import (
    AlarmRef,
)

from ...dto.summary import (
    AlarmSummary,
)

from ...dto.detail import (
    AlarmDetail,
)

from ...dto.response import (
    AlarmSummaryResponse,
)

from ...dto.payload import (
    AlarmCommentPayload,
)


class Alarms(
    APIEndpoint,
):
    """Alarms

    Provides a standard set of operations for Alarms and in addition
    allows you to follow resulting actions to act on Alarms.

    To each action you are required to pass in a Ref, Summary or Detail
    and it picks the appropriate href to follow through

    Current supported actions:
    - follow:
    - comment:
    - mark_as_viewed:
    - mark_as_acknowledged:
    - mark_as_processed:
    - mark_as_force_processed:
    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        """Return the configuration for Alarms

        Arguments:
        cls: class reference
        """
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.alarms.alarms,
            dto_list=AlarmSummaryResponse,
            dto_retrieve=AlarmDetail,
        )

    @classmethod
    async def follow(cls, href: str) -> bool:
        """ """
        return False

    @classmethod
    async def mark_as_viewed(
        cls,
        alarm: AlarmRef | AlarmSummary | AlarmDetail,
        comment: Optional[str],
    ) -> bool:
        """ Mark an alarm as viewed

        arguments:
        alarm: AlarmRef | AlarmSummary | AlarmDetail
            The alarm to mark as viewed
        comment: Optional[str]
            A comment to add to the alarm 
        """
        await cls._post(
            alarm.view.href,
            AlarmCommentPayload(comment=comment) if comment else None,
        )

        return alarm.href is not None

    @classmethod
    async def comment(
        cls,
        alarm: AlarmRef | AlarmSummary | AlarmDetail,
        comment: str,
    ) -> bool:
        """Comment on an alarm without changing status

        This will add a comment to the history of the alarm without
        changing the status of the alarm.

        You must pass a Ref, Summary or Detail which will be used
        to get the href of the Alarm

        arguments:
        alarm: AlarmRef | AlarmSummary | AlarmDetail
            The alarm to comment on
        comment: str
            The comment to add to the alarm
        """

        await cls._post(
            alarm.comment.href,
            AlarmCommentPayload(comment=comment),
        )

        return alarm.href is not None

    @classmethod
    async def mark_as_acknowledged(
        cls,
        alarm: AlarmRef | AlarmSummary | AlarmDetail,
        comment: Optional[str],
    ) -> bool:
        """ Mark an alarm as acknowledged

        arguments:
        alarm: AlarmRef | AlarmSummary | AlarmDetail
            The alarm to mark as acknowledged
        comment: Optional[str]
            A comment to add to the alarm
        """
        await cls._post(
            alarm.acknowledge.href,
            AlarmCommentPayload(comment=comment) if comment else None,
        )

        return alarm.href is not None

    @classmethod
    async def mark_as_processed(
        cls,
        alarm: AlarmRef | AlarmSummary | AlarmDetail,
        comment: Optional[str],
    ) -> bool:
        """ Mark an alarm as processed

        arguments:
        alarm: AlarmRef | AlarmSummary | AlarmDetail
            The alarm to mark as processed
        comment: Optional[str]
            A comment to add to the alarm 
        """
        await cls._post(
            alarm.process.href,
            AlarmCommentPayload(comment=comment) if comment else None,
        )

        return alarm.href is not None

    @classmethod
    async def mark_as_force_processed(
        cls,
        alarm: AlarmRef | AlarmSummary | AlarmDetail,
    ) -> bool:
        """ Mark an alarm as force processed

        arguments:
        alarm: AlarmRef | AlarmSummary | AlarmDetail
            The alarm to mark as force processed

        returns:
        bool: True if the alarm was force processed 
        """
        await cls._post(
            alarm.force_process.href,
            None,
        )

        return alarm.href is not None


__shillelagh__ = (
    Alarms,
)


""" Divisions

"""

from rich import print as rprint
from rich.console import Console
from rich.table import Table

from .utils import AsyncTyper

from gallagher.cc.alarms.divisions import Division


app = AsyncTyper(
    help="query or modify divisions"
)


@app.command("list")
async def list():
    """list all divisions"""
    console = Console()
    with console.status("[bold green]Fetching divisions...", spinner="clock"):
        divisions = await Division.list()
        table = Table(title="Divisions")
        for header in divisions.cli_header:
            table.add_column(header)

        for row in divisions.__rich_repr__():
            table.add_row(*row)

        console.print(table)


@app.command("get")
async def get(id: int):
    """get a division by id"""
    division = await Division.retrieve(id)
    [rprint(r) for r in division.__rich_repr__()]



""" Misc models

These are miscellaneous models that are used in places like alarms,
they may need to be refactored into other packages as we see fit. 
"""

from ..utils import (
    AppBaseModel,
    HrefMixin,
)


class InstructionRef(
    AppBaseModel,
    HrefMixin,
):
    """Operator reference

    At present used in Alarm details, revise as required.
    """

    name: str



""" Utilities for the Gallagher Command Centre API

This package provides a set of utilities that are used that each endpoint
uses to communicate with the Gallagher Command Centre API.

Every endpoint inherits from the APIEndpoint class and must define
a configuration that is assigned to the variable __config__.

The endpoint variable in the EndpointConfig should be assigned to a reference
to the endpoint in the DiscoveryResponse object. When initialised the
endpoint will be assigned to None but will self heal as part of 
the bootstrapping process.
"""

from typing import (
    Optional,
    Tuple,
)

from datetime import datetime
from dataclasses import dataclass

from http import HTTPStatus  # Provides constants for HTTP status codes

import httpx

from gallagher.exception import UnlicensedFeatureException

from ..dto.utils import (
    AppBaseModel,
    AppBaseResponseWithFollowModel,
)

from ..dto.detail import (
    FeaturesDetail,
)

from ..dto.response import (
    DiscoveryResponse,
)

from ..exception import (
    UnlicensedFeatureException,
    NotFoundException,
    AuthenticationError,
    DeadEndException,
    PathFollowNotSupportedError,
)


def _check_api_key_format(api_key):
    """Validates that the Gallagher Key is in the right format.

    It's not possible for the API client to validate the key against
    Gallagher servers but it will validate that the key is in the
    right format.
    """
    api_tokens = api_key.split("-")
    return api_tokens.count() == 8


def _get_authorization_headers():
    """Creates an authorization header for Gallagher API calls

    The server expects an Authorization header with GGL-API-KEY
    set to the API key provided by Gallagher Command Centre.

    This API key is what allows the cloud proxy to determine where
    the request should be routed to.

    See the Authorization section here for more information
    https://gallaghersecurity.github.io/cc-rest-docs/ref/events.html

    from gallagher import cc, const

    cc.api_key = "GH_"

    """
    from . import api_key

    return {
        "Content-Type": "application/json",
        "Authorization": f"GGL-API-KEY {api_key}",
    }


@dataclass
class EndpointConfig:
    """A configuration for an API endpoint

    Each API endpoint has a configuration that defines how
    the APIBase should attempt to parse responses from the
    Gallagher servers.

    Additionally it provides defaults for various parameters
    that are commonly accepted by endpoints.

    Not assigning a configuration to an endpoint will result
    in APIBase raising an exception.

    The configuration is assigned a special variable called
    __config__ which is set to None by default.

    """

    endpoint: str  # partial path to the endpoint e.g. day_category
    dto_list: Optional[any] = None  # DTO to be used for list requests
    dto_retrieve: Optional[any] = None  # DTO to be used for retrieve requests

    top: Optional[int] = 10  # Number of response to download
    sort: Optional[str] = "id"  # Can be set to id or -id

    fields: Tuple[str] = ()  # Optional list of fields, blank = all
    search: Tuple[str] = () # If the endpoint supports search, blank = none

    @classmethod
    async def validate_endpoint(cls):
        """Check to see if the feature is licensed and available

        Gallagher REST API is licensed per feature, if a feature is not
        the endpoint is set to none and we should throw an exception
        """
        if not cls.endpoint:
            raise UnlicensedFeatureException("Endpoint not defined")


class Capabilities:

    # Discover response object, each endpoint will reference
    # one of the instance variable Href property to get the
    # path to the endpoint.
    #
    # Gallagher recommends that the endpoints not be hardcoded
    # into the client and instead be discovered at runtime.
    #
    # Note that if a feature has not been licensed by a client
    # then the path will be set to None, if the client attempts
    # to access the endpoint then the library will throw an exception
    #
    # This value is memoized and should be performant
    CURRENT = DiscoveryResponse(
        version="0.0.0",  # Indicates that it's not been discovered
        features=FeaturesDetail(),
    )

class APIEndpoint:
    """Base class for all API objects

    All API endpoints must inherit from this class and provide a Config class
    that automates the implementation of many of the API methods.

    If the endpoints provide additional methods then they are to implement them
    based on the same standards as this base class.

    """

    # Do not set this variable in your class, this is set by the
    # lifecycle methods and use to cache the configuration object
    __config__ = None

    @classmethod
    async def expire_discovery(cls):
        """Expires endpoint discovery information

        Use this with caution as it significantly increases roundtrip times
        and decreases API client performance.

        Unless the server instance updates mid cycle, there should be no
        reason for these discovered URLs to change.
        """
        Capabilities.CURRENT = DiscoveryResponse(
            version="0.0.0",  # Indicates that it's not been discovered
            features=FeaturesDetail(),
        )

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        """Returns the configuration for the endpoint

        This method can be overridden by the child class to
        provide additional configuration options.
        """
        raise NotImplementedError("get_config method not implemented")

    @classmethod
    async def _discover(cls):
        """The Command Centre root API endpoint

        Much of Gallagher's API documentation suggests that we don't
        hard code the URL, but instead use the discovery endpoint by
        calling the root endpoint.

        This should be a singleton which is instantiated upon initialisation
        and then used across the other endpoints.

        For example features.events.events.href is the endpoint for the events
        where as features.events.events.updates is the endpoint for getting
        updates to the changes to events.

        This differs per endpoint that we work with.

        Note that references to Capabilities.CURRENT as a singleton, while
        cls.method when executing a class method.

        :params class cls: The class that is calling the method
        """

        if Capabilities.CURRENT.version != "0.0.0" and isinstance(
            Capabilities.CURRENT._good_known_since, datetime
        ):
            # We've already discovered the endpoints as per HATEOAS
            # design requirement, however because the endpoint configuration is
            # dynamically populated, we have to call the get_config method
            cls.__config__ = await cls.get_config()
            return

        # Auto-discovery of the API endpoints, this will
        # be called as part of the bootstrapping process
        from . import api_base

        async with httpx.AsyncClient() as _httpx_async:
            response = await _httpx_async.get(
                api_base,
                headers=_get_authorization_headers(),
            )

            await _httpx_async.aclose()

            parsed_obj = DiscoveryResponse.model_validate(response.json())

            # Assign the capabilities to the class, this should
            # result in the endpoint
            #
            # With the refactored initialisation of the pydantic
            # models, the values for the unavailable endpoints
            # should be set to None
            Capabilities.CURRENT = parsed_obj

            # Set this so the configuration is only discovered
            # once per endpoint
            #
            # If we assign the __config__ variable in the class
            # that inherits from this class, the instance of EndpointConfig
            # will copy the None values from the Capabilities.CURRENT
            # object, this primarily because Capabilities.CURRENT is
            # an instance of a pydantic object and all values are thus
            # copied not referenced.
            cls.__config__ = await cls.get_config()


    @classmethod
    async def list(cls, skip=0):
        """For a list of objects for the given resource

        Most resources can be searched which is exposed by this method.
        Resources also allow pagination which can be controlled by the skip

        :param int skip: fetch responses from this anchor
        """
        await cls._discover()  # Discover must be here for dynamic config

        return await cls._get(
            cls.__config__.endpoint.href,
            cls.__config__.dto_list,
        )

    @classmethod
    async def retrieve(cls, id):
        """Retrieve a single object for the given resource

        Most objects have an ID which is numeral or UUID.
        Each resource also provides a href and pagination for
        children.

        :param int id: identifier of the object to be fetched
        """
        await cls._discover()  # Discover must be here for dynamic config

        return await cls._get(
            f"{cls.__config__.endpoint.href}/{id}",
            cls.__config__.dto_retrieve,
        )

    @classmethod
    async def modify(cls):
        """ """
        pass

    @classmethod
    async def create(cls, **params):
        """ """
        cls._discover()

    @classmethod
    async def delete(cls):
        """ """
        cls._discover()

    @classmethod
    async def search(
        cls, top: int = 100, sort: str = "id", fields: str = "defaults", **kwargs
    ):
        """Search wrapper for most objects to dynamically search content

        Each object has a set of fields that you can query for, most searches
        also allow you to search for a partial string.

        :param int top: Number of results to return
        :param str sort: Sort order, can be set to id or -id
        :param str fields: List of fields to return
        :param kwargs: Fields to search for

        """
        await cls._discover()

        params = {
            "top": top,
            "sort": sort,
            "fields": fields,
        }

        # Adds arbitrary fields to the search, these will be different
        # for each type of object that calls the base function
        params.update(kwargs)

        async with httpx.AsyncClient() as _httpx_async:

            response = await _httpx_async.get(
                f"{cls.__config__.endpoint.href}",
                params=params,
                headers=_get_authorization_headers(),
            )

            await _httpx_async.aclose()

            parsed_obj = cls.__config__.dto_list.model_validate(response.json())

            return parsed_obj

    # Follow links methods, these are valid based on if the response
    # classes make available a next, previous or update href, otherwise
    # the client raises an NotImplementedError

    @classmethod
    async def next(cls, response):
        """Fetches the next set of results

        This is only valid if the response object has a next href
        """
        await cls._discover()

        # If the cls.__config__ is not of type AppBaseResponseWithFollowModel
        # then we should raise an exception
        if not issubclass(cls.__config__.dto_list, AppBaseResponseWithFollowModel):
            """A response model must have a next, previous or update"""
            raise PathFollowNotSupportedError(
                "Endpoint does not support previous, next or updates"
            )

        if not response.next:
            """We have no where to go based on the passed response"""
            raise DeadEndException("No further paths to follow for this endpoint")

        return await cls._get(
            response.next.href,
            cls.__config__.dto_list,
        )

    @classmethod
    async def previous(cls, response):
        """Fetches the previous set of results

        This is only valid if the response object has a previous href
        """
        await cls._discover()

        # If the cls.__config__ is not of type AppBaseResponseWithFollowModel
        # then we should raise an exception
        if not issubclass(cls.__config__.dto_list, AppBaseResponseWithFollowModel):
            """A response model must have a next, previous or update"""
            raise PathFollowNotSupportedError(
                "Endpoint does not support previous, next or updates"
            )

        if not response.previous:
            raise DeadEndException("No path leads further back than this")

        return await cls._get(
            cls.response.previous.href,
            cls.__config__.dto_list,
        )

    @classmethod
    async def poll(cls, response):
        """Fetches the updated set of results

        Update follow the same pattern as next and previous, except
        it keeps yielding results until the server has no more updates
        """
        await cls._discover()

        # If the cls.__config__ is not of type AppBaseResponseWithFollowModel
        # then we should raise an exception
        if not issubclass(cls.__config__.dto_list, AppBaseResponseWithFollowModel):
            """A response model must have a next, previous or update"""
            raise PathFollowNotSupportedError(
                "Endpoint does not support previous, next or updates"
            )

        return await cls._get(
            cls.response.update.href,
            cls.__config__.dto_list,
        )

    # Proposed methods for internal use
    @classmethod
    async def _get(
        cls,
        url: str,
        response_class: AppBaseModel | None,
    ):
        """Generic _get method for all endpoints

        This is to be used if we can find a prepared url endpoint, this
        is useful for follow on endpoints like commenting, next, previous
        etc. The response_class is used to parse the returned object.

        Note: that this does not run discover as it expects the endpoint to
        be fully formed and callable. If you are calling this from a generic
        endpoint like `list` please ensure you've called _discover first.

        :param str url: URL to fetch the data from
        :param AppBaseModel response_class: DTO to be used for list requests
        """
        async with httpx.AsyncClient() as _httpx_async:

            try:

                response = await _httpx_async.get(
                    f"{url}",  # required to turn pydantic object to str
                    headers=_get_authorization_headers(),
                )

                await _httpx_async.aclose()

                if response.status_code == HTTPStatus.OK:

                    if not response_class:
                        return

                    parsed_obj = response_class.model_validate(response.json())

                    return parsed_obj

                elif response.status_code == HTTPStatus.NOT_FOUND:
                    raise NotFoundException()
                elif response.status_code == HTTPStatus.FORBIDDEN:
                    raise UnlicensedFeatureException()
                elif response.status_code == HTTPStatus.UNAUTHORIZED:
                    raise AuthenticationError()

            except httpx.RequestError as e:
                pass

    @classmethod
    async def _post(
        cls,
        url: str,
        payload: AppBaseModel | None,
        response_class: AppBaseModel | None = None,
    ):
        """ """
        async with httpx.AsyncClient() as _httpx_async:

            try:

                response = await _httpx_async.post(
                    f"{url}",  # required to turn pydantic object to str
                    json=payload.model_dump() if payload else None,
                    headers=_get_authorization_headers(),
                )

                await _httpx_async.aclose()

                if response.status_code == HTTPStatus.OK:

                    if not response_class:
                        # TODO: should this return a boolean?
                        """No response to parse"""
                        return True

                    parsed_obj = response_class.model_validate(response.json())

                    return parsed_obj

                elif response.status_code == HTTPStatus.FORBIDDEN:
                    raise UnlicensedFeatureException()
                elif response.status_code == HTTPStatus.UNAUTHORIZED:
                    raise AuthenticationError()

            except httpx.RequestError as e:
                raise (e)



"""


"""

from .alarm import (
    AlarmSummaryResponse,
)

from .card_type import (
    CardTypeResponse,
)

from .cardholder import (
    CardholderSummaryResponse,
)

from .day_category import (
    DayCategoryResponse,
)

from .discover import (
    DiscoveryResponse,
)

from .division import (
    DivisionSummaryResponse,
)

from .door import (
    DoorSummaryResponse,
)

from .event import (
    EventSummaryResponse,
    EventTypeResponse,
)

from .pdf import (
    PdfResponse,
)

from .items import (
    ItemsSummaryResponse,
    ItemTypesResponse,
)

from .schedule import (
    ScheduleSummaryResponse,
)



from typing import Optional

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
    OptionalHrefMixin,
)


class DivisionDetail(AppBaseModel, IdentityMixin, HrefMixin):
    """ """

    name: str
    description: Optional[str] = None
    server_display_name: Optional[str] = None
    parent: OptionalHrefMixin = None

    # TODO: Looks like we don't have access to visitor management
    # on our test instance at the moment
    # visitor_management: visitor.VisitorManagementSummary

    def __rich_repr__(self):
        return (
            self.id,
            self.name,
            self.server_display_name \
                if self.server_display_name \
                    else "unavailable",
        )



""" Events that the Command Centre

Command Centre has about 80 event types that occur when somebody 
authenticates at a device, usually by badging a card.
"""

from ..core import Capabilities, APIEndpoint, EndpointConfig

from ...dto.detail import (
    EventDetail,
)

from ...dto.response import (
    EventTypeResponse,
    EventSummaryResponse,
)


class Event(APIEndpoint):
    """Event"""

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.events.events,
            dto_list=EventSummaryResponse,
            dto_retrieve=EventDetail,
        )


class EventType(APIEndpoint):
    """EventType"""

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.events.event_groups,
            dto_list=EventTypeResponse,
        )



""" Enumerations

This module contains enumerations used in the DTOs. Enums can be shared
across ref, detail, response, and summary DTOs.

Each one of these aligns with the API documentation and is used to
ensure that we are receive or send the correct values.
"""
from enum import Enum

class PdfType(str, Enum):
    """ Personal Data Field Types

    Types of acceptable personal data fields allowed by the server
    these are in accordance with the API endpoints
    """
    STRING = "string"
    IMAGE = "image"
    STR_ENUM = "strEnum"
    NUMERIC = "numeric"
    DATE = "date"
    ADDRESS = "address"
    PHONE = "phone"
    EMAIL = "email"
    MOBILE = "mobile"


class CredentialsClass(str, Enum):
    """ Credential Class related to different card types

    A card type must be one of the following credential classes
    """
    CARD = "card"
    DIGITAL_ID = "digitalId"
    GOV_PASS = "govPass"
    MOBILE = "mobile"
    PVI = "piv"
    PIVI = "pivi"
    TRACKING_TAG = "trackingTag"
    TRANSACT = "transact"
    FIOD2 = "fido2"


""" Extra models that may be temporary

There are some models defined in the Gallagher API that seem to be 
placeholders for features to come in future version of command centre.
"""

from typing import Optional

from ..utils import (
    AppBaseModel,
    IdentityMixin,
)

class PlaceholderRef(
    AppBaseModel,
):
    """Placeholder Ref

    This is a placeholder for a reference to a future object
    """

    reserved: Optional[str] = "for future use"



from typing import Optional

from ..utils import (
    AppBaseModel,
    IdentityMixin,
)


class EventGroupRef(
    AppBaseModel,
    IdentityMixin,
):
    """Event Group Reference

    This is a reference to an event group, it is used in the
    event type response.
    """

    name: str



""" Console
"""

from textual.app import App, ComposeResult

from textual.screen import (
    ModalScreen,
)

from textual.containers import (
    Container,
    Horizontal,
    Grid,
)

from textual.widgets import (
    Header,
    Footer,
    Placeholder,
)

from .dashboard import Dashboard

class ModalScreen(ModalScreen):

    def compose(self) -> ComposeResult:
        # Make a two column grid and occupy the second half
        # with a placeholder
        with Container():
            # yield Container()
            yield Placeholder("This is a modal screen")

class GallagherConsole(App):
    """A console interface for the Gallagher Command Centre."""

    CSS_PATH = "gallagher.tcss"

    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"), 
        ("ctrl+q", "quit", "Quit")
    ]

    # Decorative constants
    TITLE = "Gallagher"
    SUB_TITLE = "power tools for the console"

    # def on_mount(self) -> None:

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header(show_clock=True)
        yield Dashboard()
        yield Footer()

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark

    def key_m(self) -> None:
        """Open the modal screen."""
        self.push_screen(ModalScreen())


# Move this out side of __main__ so that textual cli
# and python can find them as instance variables
#
# see service.tom and you can run this via the terminal
# textual-web --config serve.toml 
# to get a web interface, follow textual-web for further
# development and how we can use it provide a demo
app = GallagherConsole()

def main():
    app.run()

if __name__ == "__main__":
    main()



""" Cardholder cli commands mounted at ch

"""

import typer

from rich.console import Console
from rich.table import Table

from .utils import AsyncTyper

from gallagher.cc.cardholders import Cardholder

from gallagher.exception import (
    NotFoundException,
)

app = AsyncTyper(
    help="query or manage cardholders"
)


@app.command("list")
async def list():
    """list all cardholders"""
    console = Console()
    with console.status("[bold green]Fetching cardholders...", spinner="dots"):

        cardholders = await Cardholder.list()

        table = Table(title="Cardholders")
        for header in cardholders.cli_header:
            table.add_column(header)

        for row in cardholders.__rich_repr__():
            table.add_row(*row)

        console.print(table)


@app.command("get")
async def get(id: int):
    """get a cardholder by id"""
    console = Console()
    with console.status("[bold]Finding cardholder...", spinner="dots"):

        try:
            cardholder = await Cardholder.retrieve(id)
            [console.print(r) for r in cardholder.__rich_repr__()]
        except NotFoundException as e:
            console.print(f"[bold]No cardholder with id={id} found[/bold]")
            raise typer.Exit(code=1)



from ..utils import (
    AppBaseModel,
)

from ..ref import (
    ScheduleRef,
    SaltoItemRef,
    SaltoItemTypeRef,
)


class SaltoAccessItemSummary(
    AppBaseModel,
):
    """A Summary of Salto items"""

    salto_item_type: SaltoItemTypeRef
    salto_item: SaltoItemRef
    schedule: ScheduleRef



from typing import Optional

from ..utils import (
    HrefMixin,
    AppBaseResponseModel,
    AppBaseResponseWithFollowModel,
)

from ..summary import ItemSummary, ItemTypeSummary


class ItemsSummaryResponse(
    AppBaseResponseWithFollowModel,
):
    """ItemsResponse is the list of items from the API
    it provides the summary of all Items Summary
    """

    results: list[ItemSummary]


class ItemTypesResponse(
    AppBaseResponseModel,
):
    """Every security centre can provide a list of item types

    While the response is rather abridged, this is the detail form
    of the Item Types.
    """

    item_types: list[ItemTypeSummary]



""" Data Transfer Object (DTO) utilities

These are utilities that DTO refs, summaries and detail objects
inherit from. It defines the basic parsing mechanism based on the
configuration of each endpoint class.

In addition we provide a based class for Responses to encapsulate
behaviour outlined by the Gallagher API.

Mixins are classes that provided an isolated behaviour for DTO
classes to inherit from, for example Identifier, Hrefs, etc.

There are some utility functions that are used internally, please
ignore them unless you are developing the client itself.

"""

from typing import (
    Optional,
)
from functools import cached_property
from typing_extensions import Annotated

from datetime import datetime

from pydantic import (
    BaseModel,
    ConfigDict,
    HttpUrl,
    Field,
)

# Annotations for fields that are reserved words in Python
from_optional_datetime = Annotated[
    Optional[datetime],
    Field(..., alias="from")
]

until_optional_datetime = Annotated[
    Optional[datetime],
    Field(..., alias="until")
]


# Helper functions for parsing
def _to_lower_camel(name: str) -> str:
    """Converts a snake_case string to lowerCamelCase

    Not designed for use outside of the scope of this package
    """
    upper = "".join(word.capitalize() for word in name.split("_"))
    return upper[:1].lower() + upper[1:]


# Ensure that the primitive wrappers such as Mixins appear
# before the generic classes for parsing utilities


class IdentityMixin(BaseModel):
    """Identifier

    This mixin is used to define the identifier field for all
    responses from the Gallagher API.
    """

    id: str


class HrefMixin(BaseModel):
    """Href

    This mixin is used to define the href field for all
    responses from the Gallagher API.
    """

    href: HttpUrl

class TypeValuePair(BaseModel):
    """Type Value Pair

    This class is used to represent a type value pair in the
    Gallagher system. It is used in many places to represent
    a key value pair.

    eg this appears in CardholderDetail:

    {
        "type": "active",
        "value": "Active"
    }

    """

    type: str
    value: str

class OptionalHrefMixin(BaseModel):
    """Optionally available Href

    This mixin is used to define the href field for all
    responses from the Gallagher API.

    Primarily used by the discovery endpoint, where the href
    may be absent if the feature is not available.

    Reason for this so the API Endpoint configuration can
    reference the href property (pre discovery), otherwise
    the Feature* classes have a None object for the object

    # Use with caution

    Only use these with responses that don't optionally
    require a href. See Gallagher's documentation for
    confirmation.
    """

    href: Optional[HttpUrl] = None


class AppBaseModel(BaseModel):
    """Pydantic base model for applications

    This class is used to define the base model for all schema
    that we use in the Application, it configures pydantic to
    translate between camcelCase and snake_case for the JSON
    amongst other default settings.

    ORM mode will allow pydantic to translate SQLAlchemy results
    into serializable models.

    For a full set of options, see:
    https://pydantic-docs.helpmanual.io/usage/model_config/
    """

    model_config = ConfigDict(
        populate_by_name=True,
        alias_generator=_to_lower_camel,
        from_attributes=True,
        arbitrary_types_allowed=True,
    )

    @classmethod
    def _accumulated_annotations(cls) -> dict:
        """Return a dictionary of all annotations

        This method is used to return a dictionary of all the
        annotations from itself and it's parent classes.

        It is intended for use by the shillelagh extension
        """
        annotations = cls.__annotations__.copy() # TODO: should we  make copies?
        for base in cls.__bases__:
            if issubclass(base, BaseModel):
                # Copy annotations from classes that are pydantic models
                # they are the only things that form part of the response
                annotations.update(base.__annotations__)
        return annotations.items()

    # Set to the last time each response was retrieved
    # If it's set to None then the response was either created
    # by the API client or it wasn't retrieved from the server
    #
    # This is generally used for caching
    _good_known_since: Optional[datetime] = None

    def model_post_init(self, __context) -> None:
        """
        The model_post_init method is called after the model is
        initialized, this is used to set the _good_known_since

        https://docs.pydantic.dev/2.0/api/main/#pydantic.main.BaseModel.model_post_init
        """
        self._good_known_since = datetime.now()

    def __repr__(self) -> str:
        """Return a string representation of the model

        This method is used to return the string representation of the
        model, it's used for debugging purposes.

        https://docs.python.org/3/reference/datamodel.html
        """
        return f"{self.__class__.__name__}({self.model_dump()})"


class AppBaseResponseModel(AppBaseModel):
    """Response Model

    Response classes should subclass this not AppBaseModel, so that
    the framework can differentiate between Model classes and responses.

    AppBaseResponseWithFollowModel sub classes from this, and adds
    functionality for following next, updates or going back to the previous
    set of results.

    """

    pass


class AppBaseResponseWithFollowModel(AppBaseResponseModel):
    """Response with optional Follow links

    Many responses from the Gallagher system return next, previous and update
    links. This is typical when there are a large number of responses
    where the absence of a next link denotes that the client has fetched
    all pending objects.

    Some endpoints return an `update` URL which can be followed to poll
    for responses.


    """

    next: OptionalHrefMixin = None  # None means it's the end of responses
    previous: OptionalHrefMixin = None  # None means first set of responses
    updates: OptionalHrefMixin = None  # None means no updates to watch for



""" Command Centre bindings

    After enabling the Gateway for your server and REST API user, you need 
    only to replace the local server's URL with the cloud API Gateway one.

    Example Local request:

        <client certificate thumbprint e89ef121958...>
        GET https://servername.yourcompany.local:8904/api/cardholders
        Authorization: GGL-API-KEY F6F0-C8F0-
        
    Equivalent Gateway request:

        <client certificate thumbprint e89ef121958...>
        GET https://commandcentre-api-au.security.gallagher.cloud/api/cardholders
        Authorization: GGL-API-KEY F6F0-C8F0-60AF-

    The API client takes of appending the headers onto the requests.

    The packages mimic the documentation structure presented by Gallagher on their
    Github pages https://gallaghersecurity.github.io/cc-rest-docs/ref/index.html
"""

from typing import Optional

from ..const import URL

# Follow the instructions in the Gallagher documentation
# to obtain an API key
api_key: str = None

# By default the base API is set to the Australian Gateway
# Override this with the US gateway or a local DNS/IP address
api_base: str = URL.CLOUD_GATEWAY_AU

# Default is set to the library, set this to your application
client_id: str = "gallagher-py"

# By default connections are sent straight to the server
# should you wish to use a proxy, set this to the proxy URL
proxy: Optional[str] = None



"""

"""

from ..core import Capabilities, APIEndpoint, EndpointConfig

from ...dto.response import CardTypeResponse


class CardType(APIEndpoint):
    """Card Types provide a list of support card types for the instance.

    These can vary between using physical cards, mobile credentials, or
    biometrics. The card type is used to dynamically determine the types
    of credentials available on this particular instance.
    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.card_types.card_types,
            dto_list=CardTypeResponse,
            dto_retrieve=CardTypeResponse,
        )



from ..utils import AppBaseModel, HrefMixin


class SaltoItemTypeRef(
    AppBaseModel,
):
    """A Salto Item Type is a SALTO product category e.g door"""

    value: str


class SaltoItemRef(AppBaseModel, HrefMixin):
    """A Salto Item is a particular product e.g Salto CU5000"""

    name: str



"""

"""

from ..core import Capabilities, APIEndpoint, EndpointConfig

from ...dto.response import ScheduleSummaryResponse


class Schedule(APIEndpoint):
    """Schedules"""

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.schedules.schedules,
            dto_list=ScheduleSummaryResponse,
        )



from typing import Optional

from datetime import datetime

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
    from_optional_datetime,
    until_optional_datetime,
)

from ..ref import (
    CardTypeRef,
)

from ..enum import CredentialsClass

class CardholderSummary(
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
):
    """
    The cardholder search at /api/cardholders returns an array of these.
    It is a subset of what you get from a cardholder's detail page at
    /api/cardholders/{id}

    (linked as the href in this object), to be more suitable for large result sets.
    """

    first_name: str
    last_name: str
    short_name: Optional[str] = None
    description: Optional[str] = None
    authorised: bool = False

    def __rich_repr__(self):
        return (
            self.id,
            self.first_name,
            self.last_name,
            "yes" if self.authorised else "no",
        )

    def __str__(self):
        return f"{self.id} {self.first_name} {self.last_name}"


class CardholderCardStatusSummary(
    AppBaseModel,
):
    """ Used by CardholderCardSummary to show the status of an assigned card
    """
    value: str
    type: str


class CardholderCardInvitationSummary(
    AppBaseModel,
    HrefMixin,
):
    """ Used by CardholderCardSummary to show status of an invitation

    Invitations are sent by the Gallagher system to a cardholder via
    email and then text messages for mobile credentials to activate
    the app.
    """
    email: str
    mobile: Optional[str] = None
    single_factor_only: bool = False
    status: str

class CardholderCardSummary(
    AppBaseModel,
    HrefMixin,
):
    """ A card assigned to a cardholder, note that these are not the same as
    the card objects themselves, they miss a few fields that the card object
    has and are only used in the context of a cardholder's detail page.
    
    These children objects are referenced by their `href` as they aren't assigned
    an identifier object in the system.
    """
    number: str
    card_serial_number: Optional[str] = None
    issue_level: Optional[int] = None

    status: CardholderCardStatusSummary

    type: CardTypeRef

    invitation: Optional[CardholderCardInvitationSummary] = None

    valid_from: from_optional_datetime = None # Appears as from in the API
    valid_until: until_optional_datetime = None
    credential_class: CredentialsClass
    trace: bool = False
    last_printed_or_encoded_time: Optional[datetime] = None
    last_printed_or_encoded_issue_level: int = 0
    pin: Optional[str] = None # Leading zeros are significant
    visitor_contractor: bool = False
    owned_by_site: bool = False
    credential_id: str = "reserved" # TODO: Reserved for future use?
    # TODO: Reserved for future use?
    # Test instance returns an int when a mobile credential is assigned
    # the docs still suggest this should be a string instead, leaving this
    # here to check into the future
    ble_facility_id:  str | int = "reserved" 




"""
"""

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
    OptionalHrefMixin,
)

class CardTypeRef(
    AppBaseModel,
    HrefMixin,
):
    """ Reference for a card type """
    name: str


from typing import Optional

from ..utils import (
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
    OptionalHrefMixin,
)


class CardholderRef(
    AppBaseModel,
    HrefMixin,
):
    """Reference to a Cardholder"""

    name: str


class CardholderExtendedRef(
    AppBaseModel,
    IdentityMixin,
    OptionalHrefMixin,
):
    """Cardholder ref used with events

    Events seem to send partial information where href is missing at times
    (this is contrary to the documentation), name and id seem to always
    be present. This is likely because this is a summary
    """

    name: str
    first_name: Optional[str] = None
    last_name: Optional[str] = None



from typing import Optional
from datetime import datetime

from ..utils import (
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
    OptionalHrefMixin,
)

from ..ref import (
    AlarmRef,
    CardholderRef,
    CardholderExtendedRef,
    DoorRef,
    AccessZoneRef,
    DivisionRef,
    ItemRef,
    EventGroupRef,
)

from .card_type import (
    CardSummary,
)

from .items import (
    ItemSummary,
)


class EventTypeSummary(
    AppBaseModel,
    IdentityMixin,
):
    """An event type has identifiers and names"""

    name: str


class EventGroupSummary(
    AppBaseModel,
    IdentityMixin,
):
    """Event Groups are a collection of event types

    Each group has names and event types. This is usually used
    in an Event Type Response.
    """

    name: str
    event_types: list[EventTypeSummary]


class EventSummary(
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
):
    """Summary of events that have occurred on the server"""

    server_display_name: Optional[str] = None
    time: datetime
    message: Optional[str] = None
    occurrences: Optional[int] = 0
    priority: int
    alarm: Optional[AlarmRef] = None

    operator: Optional[CardholderRef] = None
    source: ItemRef
    group: Optional[EventGroupRef] = None
    type: Optional[EventTypeSummary] = None
    event_type: Optional[EventTypeSummary] = None
    division: Optional[DivisionRef] = None
    cardholder: Optional[CardholderExtendedRef] = None
    entry_access_zone: Optional[AccessZoneRef] = None
    exit_access_zone: Optional[AccessZoneRef] = None
    door: Optional[DoorRef] = None
    access_group: OptionalHrefMixin = None
    card: Optional[CardSummary] = None
    modified_item: Optional[ItemSummary] = None



from ..utils import AppBaseModel, IdentityMixin

from ..ref import AccessGroupRef

from .access_group import (
    AccessGroupSummary,
)


class VisitorTypeSummary(AppBaseModel, IdentityMixin):
    """Visitor Types are a combination of:
    - Access Group
    - Host Access Groups
    - Visitor Access Groups

    these are represented in Divisions
    """

    access_group: AccessGroupRef
    host_access_groups: list[AccessGroupSummary]
    visitor_access_groups: list[AccessGroupSummary]


class VisitorManagementSummary(AppBaseModel):
    """This is the summary of the Visitor Management that appears
    in Divisions.
    """

    active: bool
    visitor_types: list[VisitorTypeSummary]



from ..utils import AppBaseModel, IdentityMixin, HrefMixin


class DivisionRef(AppBaseModel, IdentityMixin, HrefMixin):
    """Division reference is used to link to a division

    The Mixins cover all the fields that are returned in the
    summary, hence nothing has to be declared in the body
    """

    pass



"""


"""

from ..core import Capabilities, APIEndpoint, EndpointConfig

from ...dto.response import DayCategoryResponse


class DayCategory(APIEndpoint):
    """Day Categories"""

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.day_categories.day_categories,
            dto_list=DayCategoryResponse,
            dto_retrieve=DayCategoryResponse,
        )



""" Extensions package for building upon other frameworks

This package provides extensions around framewokrs like shillelagh
that allow us to extend functionality like a SQL interface to
the package.


It's recommdend that we list the package that ext provides for 
developers to understand what is possible and in what context.

- [shillelagh](https://youtu.be/85ztwhfyQ-k?feature=shared&t=14) 
  how to pronounce

"""



"""


"""

from .access_group import (
    AccessGroupDetail,
)

from .alarm import (
    AlarmDetail,
)

from .card_type import (
    CardTypeDetail,
)

from .cardholder import (
    CardholderDetail,
)

from .discover import (
    FeaturesDetail,
)

from .division import (
    DivisionDetail,
)

from .event import (
    EventDetail,
)

from .pdf import (
    PdfDetail,
)

from .role import (
    RoleDetail,
)



from typing import Optional
from datetime import datetime

from ..utils import (
    AppBaseModel,
    HrefMixin,
    OptionalHrefMixin,
    IdentityMixin,
)

from .event import EventTypeSummary


class AlarmSourceSummary(
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
):
    """AlarmSource represents a device that has triggered an alarm"""

    name: str


class AlarmSummary(
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
):
    """AlarmSummary gives us unactioned events from the CC

    While the detail and summary would typically differ, it seems that
    the API endpoints return more detail in the summary endpoints.

    Note: that we have a number of hrefs that we should follow to
    HATEOAS compliance as per the documentation.

    """

    time: datetime
    message: str
    source: AlarmSourceSummary
    type: str
    event_type: Optional[EventTypeSummary] = None
    priority: int
    state: str
    active: bool
    division: HrefMixin
    event: OptionalHrefMixin = None
    note_presets: list[str] = []

    # The following URLS should be used to follow through
    # on various actions that the system allows
    view: HrefMixin
    comment: HrefMixin
    acknowledge: OptionalHrefMixin = None
    acknowledge_with_comment: OptionalHrefMixin = None
    process: OptionalHrefMixin = None
    process_with_comment: OptionalHrefMixin = None
    force_process: OptionalHrefMixin = None

    def __rich_repr__(self):
        return (
            self.id,
            f"{self.priority}",
            self.time.strftime('%c'),
            self.type,
        )

    def __str__(self):
        return f"{self.id} {self.priority} {self.type}"




from typing import Optional

from ..utils import (
    AppBaseModel,
    HrefMixin,
)

from ..ref import (
    DivisionRef,
    AccessGroupRef,
)

from ..enum import PdfType

class PdfDetail(AppBaseModel, HrefMixin):
    """Personal Data Fields are custom fields for a card holder"""

    name: str
    server_display_name: Optional[str] = None
    description: Optional[str] = None
    division: DivisionRef
    type: PdfType
    default: Optional[str] = None
    required: bool = False
    unique: bool = False
    default_access: str
    operator_access: Optional[str] = None
    sort_priority: int

    access_groups: list[AccessGroupRef]

    regex: Optional[str] = None
    regex_description: Optional[str] = None
    notification_default: bool = False
    image_width: int = 0
    image_height: int = 0
    image_format: Optional[str] = None
    content_type: Optional[str] = None
    is_profile_image: bool = False



from typing import (
    Annotated,
)

from ..utils import (
    AppBaseResponseModel,
)

from ..detail import (
    FeaturesDetail,
)


class DiscoveryResponse(
    AppBaseResponseModel,
):
    """A response that outlines the capability of the server

    Gallagher requires customers to license individual features, if they are
    the server will return a 403 HTTP code. The purpose of this model is to
    discover what features are available on the server.

    The response should be memoized as it is unlikely to change during individual
    sessions, they can however change over a period of time.

    This API client is updated to work with various versions of the server, the
    server responds with a version string that can be used to determine if
    the API client can work with the server.
    """

    version: Annotated[str, "The version of the server"] = "0.0.0"
    features: Annotated[
        FeaturesDetail, "A list of features available on the server"
    ] = FeaturesDetail()

    @property
    async def get_sem_ver(self):
        """Get a SemVer tuple from the version string"""
        return self.version.split(".")



""" Personal Data Fields

Custom fields defined per command centre installation, these need
to be discovered ahead of time before we work with them.

Note:
- Keys are prefixed with @
- Presence of space in the field name
"""

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
)


class PDFRef(AppBaseModel, IdentityMixin, HrefMixin):
    """Personal Data Fields are custom fields for a card holder

    These are defined per command centre installation and the fields
    are prefixed with @ e.g @Personal URL, note the presence of the space
    in the field name.

    Since they are dynamic we will have to discover the personal data
    fields much like the URL discovery before we are able to parse the data
    """

    name: str



from typing import Optional

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
    OptionalHrefMixin,
    TypeValuePair,
    from_optional_datetime,
    until_optional_datetime,
)

from ..ref import (
    AccessGroupRef,
    AccessZoneRef,
    ScheduleRef,
)


class AccessSummary(
    AppBaseModel,
    HrefMixin,
):
    """Access is zone paired with a schedule"""

    access_zone: AccessZoneRef
    schedule: ScheduleRef


class AccessGroupSummary(
    AppBaseModel,
):
    """AccessGroup Summary is what the API returns on searches

    This builds on the Ref class to add the summary fields and is
    extended by the Detail class to add the fully remainder of
    the fields
    """

    name: str
    description: Optional[str]
    parent: Optional[AccessGroupRef]
    division: IdentityMixin
    cardholders: OptionalHrefMixin
    server_display_name: Optional[str]



class CardholderAccessGroupSummary(
    AppBaseModel,
    HrefMixin,
):
    """ An Access Group assigned to a cardholder.

    Essentially when an access group is assigned to a cardholder
    it inherits the properties and additionally a start and end date
    """
    access_group: AccessGroupRef
    status: TypeValuePair

    valid_from: from_optional_datetime = None # Appears as from in the API
    valid_until: until_optional_datetime = None



from typing import Optional

from ..utils import AppBaseModel, IdentityMixin, HrefMixin


class ItemTypeSummary(
    AppBaseModel,
):
    """Items Types only provide the name and id

    This is used by the discovery endpoint and is used by
    the ItemTypesResponse
    """

    id: str
    name: str
    canonical_type_name: Optional[str] = None


class ItemSummary(
    AppBaseModel,
    HrefMixin,
    IdentityMixin,
):
    """Summary of an Item which adds the notes and
    server_display_name, this is used by the item summary response
    """

    name: Optional[str] = None
    type: ItemTypeSummary



""" Data Transfer Object

This package defines the data transfer objects (DTOs) to parse
the data from the Gallagher API into Python objects.

We use pyndatic to define the DTOs. pydantic is able to validate
the data types and values of the DTOs.

Our ultimate aim is define rules as outlined by their documentation.

Resources:
  - https://bit.ly/3UkhQhS
  
"""



""" Constants for the Gallagher Cloud ecosystem 

Gallagher publishes a set of networking constants for their cloud gateways
these should be updated as the documentation is updated.
"""


class URL:
    """DNS names for the cloud gateways

    These are published by Gallagher on Github
    https://gallaghersecurity.github.io/commandcentre-cloud-api-gateway.html
    """

    CLOUD_GATEWAY_AU: str = "https://commandcentre-api-au.security.gallagher.cloud/api/"
    CLOUD_GATEWAY_US: str = "https://commandcentre-api-us.security.gallagher.cloud/api/"


class IP_ADDR:
    """IP addresses for the cloud gateways

    These are published by Gallagher on Github
    https://gallaghersecurity.github.io/commandcentre-cloud-api-gateway.html
    """

    CLOUD_GATEWAY_AU = ["3.106.1.6", "3.106.100.112"]

    CLOUD_GATEWAY_US = ["44.193.42.111", "3.209.194.103"]



from ..utils import (
    AppBaseModel, 
    OptionalHrefMixin,
)


class AccessGroupRef(AppBaseModel, OptionalHrefMixin,):
    """Access Groups is what a user is assigned to to provide access to doors"""

    name: str



""" Contains Enumerations for the Gallagher objects

Our naming convention is ObjectIntent

CustomerSearch

"""

from enum import Enum


class CustomerSort(Enum):
    """Sort descriptors for the Customer object"""

    ID: str = "id"
    ID_DESC: str = "-id"

    NAME: str = "name"
    NAME_DESC: str = "-name"



""" A day category links a calendar to a schedule. 

The calendar determines the days of the year that fall into a day category, 
and the schedule determines what happens at certain times on those days.
"""

from ..utils import AppBaseModel, HrefMixin


class DayCategoryRef(
    AppBaseModel,
    HrefMixin,
):
    """A reference to a day category

    This is what is sent by the day_category endpoint as of v9
    the references can be used from other endpoints.
    """

    name: str



""" Items


"""

from ..core import (
    Capabilities,
    APIEndpoint,
    EndpointConfig,
)

from ...dto.summary import (
    ItemSummary,
)
from ...dto.response import (
    ItemTypesResponse,
    ItemsSummaryResponse,
)


class ItemsTypes(APIEndpoint):
    """
    Gallagher
    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.items.item_types,
            dto_list=ItemTypesResponse,
            dto_retrieve=ItemTypesResponse,
        )


class Item(APIEndpoint):
    """
    Gallagher advises against hardcoding the URLs for divisions, and instead
    recommends using the /api endpoint to discover the URLs from
    events.divisions.href and alarms.division.href.

    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.items.items,
            dto_list=ItemsSummaryResponse,
            dto_retrieve=ItemSummary,
        )



""" Exceptions declared by the Gallagher parser package

These are exceptions that extend standard Python exceptions to
indicate critical issues with configuration or parsing of
responses from the Gallagher API.

Everything declared here indicates an implementation error.
"""


class GCCBaseException(Exception):
    """Base exception for all Gallagher exceptions

    This is the base exception for all Gallagher exceptions.
    """


class UnlicensedFeatureException(GCCBaseException):
    """Raised when a feature is not licensed

    This exception is raised when the client attempts to access
    an endpoint that is not licensed by the server.

    """


class NotFoundException(GCCBaseException):
    """Raised if you tried to access an object that either does not
    exists or you don't have permission to access it.
    """


class ComingSoonException(GCCBaseException):
    """Raised if Gallagher has marked this feature to be coming soon

    These are items such as mutating a division which is scheduled
    but has not be released
    """


class DeadEndException(GCCBaseException):
    """Raised if an API the path is a dead end

    This is raised if the API path is not supported by the client
    these related to next, previous or updates that certain
    endpoints support.
    """


class AuthenticationError(GCCBaseException):
    """Error authentication against the CC API

    This is likely because the use has not provided an authentication
    key or is not allowed to access this resources with this
    particular key
    """


class PathFollowNotSupportedError(GCCBaseException):
    """Raised if the path is not supported by the client

    This is raised if the API path is not supported by the client
    these related to next, previous or updates that certain
    endpoints support.
    """



""" Locker groups """

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
    OptionalHrefMixin,
)


class CardholderLockerSummary(
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
):
    """
    """
    pass




from typing import Optional

from ..utils import (
    AppBaseModel,
    IdentityMixin,
)

class CardExpiryTypeSummary(AppBaseModel):
    expiry_type: Optional[str] = None


class CardSummary(
    AppBaseModel,
    IdentityMixin,
):
    """Card summary as sent by the Event objects

    Note: that we should revise this if required

    """

    facility_code: str
    number: str
    issue_level: int



""" Alarms CLI

Sub commands to interact with the alarms data

"""

from typing import Optional
from typing_extensions import Annotated

import typer

from rich.console import Console
from rich.progress import Progress, TaskID
from rich.table import Table

from .utils import AsyncTyper

from ..exception import (
    NotFoundException,
)

from gallagher.cc.alarms import (
    Alarms,
)

app = AsyncTyper(
    help="list, query, follow, act on alarms in the command centre"
)


@app.command("list")
async def list():
    """list current alarms"""

    console = Console()

    with console.status("[bold green]Fetching alarms...", spinner="dots"):

        alarms = await Alarms.list()
        table = Table(title="Alarms")

        for header in alarms.cli_header:
            table.add_column(header)

        for row in alarms.__rich_repr__():
            table.add_row(*row)

        console.print(table)


@app.command("get")
async def get(
    id: Annotated[int, typer.Argument(help="alarm id")],
):
    """get alarm details
    
    These are details of the alarm that are given by the command centre\n
    - Item one\n
    - Items two
    """
    console = Console()
    with console.status("[bold green]Fetching alarm...", spinner="dots"):

        try:
            alarm = await Alarms.retrieve(id)
            [console.print(r) for r in alarm.__rich_repr__()]
        except NotFoundException as e:
            console.print(f"[bold]No alarm with id={id} found[/bold]")
            raise typer.Exit(code=1)

@app.command("history")
async def history(
    id: Annotated[int, typer.Argument(help="alarm id")],
):
    """show history of an alarm"""
    console = Console()

    raise NotImplementedError("Not implemented")

@app.command("tail")
async def tail():
    """watch for alarm updates"""
    console = Console()

    raise NotImplementedError("Not implemented")


@app.command("comment")
async def comment(
    id: Annotated[int, typer.Argument(help="alarm id")],
    message: Annotated[
        str,
        typer.Option(
            "-m",
            "--message",
            help="comment to add to history",
        ),
    ],
):
    """comment on an alarm"""
    console = Console()
    with console.status(
        "[magenta] Commenting on alarm ...",
    ) as status:
        try:
            console.log("Finding alarm ...")
            comment_detail = await Alarms.retrieve(id)
            console.log("Adding comment to history ...")
            await Alarms.comment(comment_detail, message)
            console.print("[green]Comment posted successfully[/green]")
        except NotFoundException as e:
            console.print(f"[red bold]No alarm with id={id} found")
            raise typer.Exit(code=1)


@app.command("ack")
async def acknowledge(
    id: Annotated[int, typer.Argument(help="alarm id")],
    message: Annotated[
        Optional[str],
        typer.Option(
            "-m",
            "--message",
            help="comment to add to history",
        ),
    ] = None,
):
    """acknowledge an alarm, optionally with a comment"""
    console = Console()


@app.command("view")
async def acknowledge(
    id: Annotated[int, typer.Argument(help="alarm id")],
    message: Annotated[
        Optional[str],
        typer.Option(
            "-m",
            "--message",
            help="comment to add to history",
        ),
    ] = None,
):
    """mark alarm as viewed, optionally with a comment"""
    console = Console()


@app.command("process")
async def acknowledge(
    id: Annotated[int, typer.Argument(help="alarm id")],
    message: Annotated[
        Optional[str],
        typer.Option(
            "-m",
            "--message",
            help="comment to add to history",
        ),
    ] = None,
):
    """mark alarm as processed, optionally with a comment"""
    console = Console()



""" Personal Data Fields

Dynamically occurring fields are defined per site and are
prefixed by an @ symbol as the key of the field, as they
appear in the API responses.

These fields appear either under the personalDataFields as
href references or as individual fields in the user's profile.

"""

from ..utils import (
    AppBaseResponseModel,
)

from ..ref import (
    PDFRef,
)


class PdfResponse(
    AppBaseResponseModel,
):
    """Personal Definition fields

    Returned as a set of results, each result is a PDFRef
    these are to be cached just as we do the URL endpoints,
    note that this must be done after the discovery completes
    """

    results: list[PDFRef]



""" Command Centre API discovery

The Command Centre API has a discovery endpoint that allows

"""

from ..utils import (
    AppBaseModel,
    OptionalHrefMixin,
)


class FeatureAccessGroupsRef(
    AppBaseModel,
):
    access_groups: OptionalHrefMixin = None


class FeatureAccessZonesRef(
    AppBaseModel,
):
    access_zones: OptionalHrefMixin = None


class FeatureAlarmZonesRef(
    AppBaseModel,
):
    alarm_zones: OptionalHrefMixin = None


class FeatureAlarmsRef(
    AppBaseModel,
):
    alarms: OptionalHrefMixin = None
    divisions: OptionalHrefMixin = None
    updates: OptionalHrefMixin = None


class FeatureCardTypesRef(
    AppBaseModel,
):
    assign: OptionalHrefMixin = None
    card_types: OptionalHrefMixin = None


class FeatureCardholdersRef(
    AppBaseModel,
):
    cardholders: OptionalHrefMixin = None
    changes: OptionalHrefMixin = None
    update_location_access_zones: OptionalHrefMixin = None


class FeatureCompetenciesRef(
    AppBaseModel,
):
    competencies: OptionalHrefMixin = None


class FeatureDayCategoriesRef(
    AppBaseModel,
):
    day_categories: OptionalHrefMixin = None


class FeatureDivisionsRef(
    AppBaseModel,
):
    divisions: OptionalHrefMixin = None


class FeatureDoorsRef(
    AppBaseModel,
):
    doors: OptionalHrefMixin = None


class FeatureElevatorsRef(
    AppBaseModel,
):
    elevator_groups: OptionalHrefMixin = None


class FeatureEventsRef(
    AppBaseModel,
):
    divisions: OptionalHrefMixin = None
    event_groups: OptionalHrefMixin = None
    events: OptionalHrefMixin = None
    updates: OptionalHrefMixin = None


class FeatureFenceZonesRef(
    AppBaseModel,
):
    fence_zones: OptionalHrefMixin = None


class FeatureInputsRef(
    AppBaseModel,
):
    inputs: OptionalHrefMixin = None


class FeatureInterlockGroupsRef(
    AppBaseModel,
):
    interlock_groups: OptionalHrefMixin = None


class FeatureItemsRef(
    AppBaseModel,
):
    item_types: OptionalHrefMixin = None
    items: OptionalHrefMixin = None
    updates: OptionalHrefMixin = None


class FeatureLockerBanksRef(
    AppBaseModel,
):
    locker_banks: OptionalHrefMixin = None


class FeatureMacrosRef(
    AppBaseModel,
):
    macros: OptionalHrefMixin = None


class FeatureOperatorGroupsRef(
    AppBaseModel,
):
    operator_groups: OptionalHrefMixin = None


class FeatureOutputsRef(
    AppBaseModel,
):
    outputs: OptionalHrefMixin = None


class FeaturePersonalDataFieldsRef(
    AppBaseModel,
):
    personal_data_fields: OptionalHrefMixin = None


class FeatureReceptionsRef(
    AppBaseModel,
):
    receptions: OptionalHrefMixin = None


class FeatureRolesRef(
    AppBaseModel,
):
    roles: OptionalHrefMixin = None


class FeatureSchedulesRef(
    AppBaseModel,
):
    schedules: OptionalHrefMixin = None


class FeatureVisitsRef(
    AppBaseModel,
):
    visits: OptionalHrefMixin = None



from typing import Optional

from ..utils import AppBaseModel, IdentityMixin, HrefMixin

from ..ref import DivisionRef


class RoleDetail(AppBaseModel, IdentityMixin, HrefMixin):
    """ """

    name: str
    server_display_name: str
    description: Optional[str]
    division: DivisionRef



""" Operator

An operator is someone who has operated an door and caused an
event like an alarm.

"""

from ..utils import (
    AppBaseModel,
)


class OperatorRef(
    AppBaseModel,
):
    """Operator reference

    At present used in Alarm details, revsie as required.
    """

    name: str



""" Cardholder Detail """
from typing import Optional, Any
from typing_extensions import Self

from pydantic import model_validator, Field

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
)

from ..ref import (
    CardholderExtendedRef,
    DivisionRef,
    PlaceholderRef,
    RoleRef,
)

from ..summary import (
    CardholderCardSummary,
    CardholderAccessGroupSummary,
    PdfSummary,
)

class CardholderRelationshipDetail(
    AppBaseModel,
    HrefMixin,
):
    """ A role that a cardholder has 

    Think of this as a representation of a many to many relationship
    where this cardholder has a role in the system. Other users can
    have the same role.
    """
    role: RoleRef
    cardholder: CardholderExtendedRef


class CardholderPersonalDataField(
    AppBaseModel,
    HrefMixin,
):
    """ A PDF field as defined in the Cardholders personal data

    A definition that defines the cardholder's personal data. This
    is essentially the Summary of the Pdf field, along with a string
    value, which also is accessible using the key at the dictionary level.
    """
    definition: PdfSummary
    value: str | HrefMixin
    notifications: Optional[bool] = False # Local to the @Email field

class CardholderPersonalDataDefinition(
    AppBaseModel,
):
    """ A personal data definition for the cardholder

    This is sent back as the personalDataDefinitions field in the
    CardholderDetail response. It has a peculiar structure that 
    we have outlined in the parse_personal_data_definitions method
    """
    name: str
    contents: CardholderPersonalDataField


class PdfAccessorWrapper:
    """ Wrapper that holds a dictionary of accessible PDF 

    The aim is to make these accessible in as python a way as possible

    Note: this is not a pydantic class, and is used as a placeholder
    to dynamically populate the dictionary of personal data fields
    """
    pass
    
class CardholderDetail(
    AppBaseModel,
    IdentityMixin,
):
    """Displays a table of cardholders

    Gallagher command centre offers a summary of all the cardholders
    provisioned on the system. This command presents a summary table
    with the aim of using the identifier to get detailed information.

    """

    first_name: str
    last_name: str
    short_name: Optional[str] = None
    description: Optional[str] = None
    authorised: bool

    disable_cipher_pad: bool = False
    division: DivisionRef
    edit: HrefMixin

    operator_login_enabled: bool = False
    operator_password_expired: bool = False

    update_location: HrefMixin
    updates: HrefMixin

    user_extended_access_time: bool = False
    windows_login_enabled: bool = False

    personal_data_definitions: list[CardholderPersonalDataDefinition] = []
    cards: list[CardholderCardSummary] = []
    access_groups: list[CardholderAccessGroupSummary] = []
    # operator_groups
    # competencies

    edit: HrefMixin
    update_location: HrefMixin
    notes: Optional[str] = None

    # notifications
    relationships: list[CardholderRelationshipDetail] = []
    # lockers
    # elevator_groups
    updates: PlaceholderRef
    # redactions

    # Note this is not a pyndatic class and hence utils.py
    # has the configuration set to allow arbitrary types
    pdf: PdfAccessorWrapper = PdfAccessorWrapper()

    @model_validator(mode='before')
    @classmethod
    def parse_personal_data_definitions(cls, data: Any) -> Any:
        """ Rewrite the personalDataDefinition for it to be parseable

        The personalDataDefinitions field is a list of objects, each one of 
        which has a single object with key being the personal data field name
        and the value being the value of the field.

        This method rewrites that to be an object with name and contents fields
        which allows us to use pydantic to parse it and make it available to the
        user as a list of objects.

        The results of this will also be used to dynamically parse the values
        for the cardholder's PDF fields.
        """

        if 'personalDataDefinitions' in data:
            data['personalDataDefinitions'] = [
                {
                    'name': name, 
                    'contents': contents
                } \
                    for item in data['personalDataDefinitions'] \
                    for name, contents in item.items()
            ]

        return data
    
    @model_validator(mode='after')
    def populate_pdf_accessor(self) -> Self:
        # For each key in the personal_data_definitions, create a new
        # attribute on the pdf object with the key as the name
        for pdf_field in self.personal_data_definitions:
            # Note this sets a reference from personal_data_definitions
            # if you compare these in the tests, they should be the same
            setattr(
                self.pdf, 
                # replace spaces with underscores and make it lowercase
                # ignore the prefixed @ symbol
                pdf_field.name[1:].replace(' ', '_').lower(),
                pdf_field.contents
            )

        return self

    def __rich_repr__(self):
        return (
            f"[blue bold] person",
            f"{'id':>20} {self.id}",
            f"{'first_name':>20} {self.first_name}",
            f"{'last_name':>20} {self.last_name}",
            f"{'short_name':>20} {self.short_name}",
            f"{'description':>20} {self.description}",
            f"{'authorised':>20} {'yes' if self.authorised else 'no'}",
            f"",
            f"{'disable_cipher_pad':>20} {'yes' if self.disable_cipher_pad else 'no'}",
            f"{'division':>20} {self.division.id}",
            f"[blue bold] hrefs",
            f"{'edit':>20} [link={self.edit.href}]edit[/link]",
        )



""" Division Summary

"""

from typing import Optional

from ..utils import (
    AppBaseModel,
    IdentityMixin,
    HrefMixin,
    OptionalHrefMixin,
)


class DivisionSummary(AppBaseModel, IdentityMixin, HrefMixin):
    """Division Summary

    Primarily separated for the cli to be able to distinguish between
    sumamry and detail responses, this is also very useful for the CLI
    """

    name: str
    description: Optional[str] = None
    server_display_name: Optional[str] = None
    parent: OptionalHrefMixin = None

    def __rich_repr__(self):
        return (
            self.id,
            self.name,
            self.server_display_name \
                if self.server_display_name \
                else "unavailable",
        )



from ..utils import (
    AppBaseModel,
    HrefMixin,
)


class AlarmRef(AppBaseModel, HrefMixin):
    """AlarmRef represents a single alarm"""

    state: str


class AlarmZoneRef(AppBaseModel, HrefMixin):
    """AccessZone represents"""

    name: str



""" CLI entry point

gal is the gallagher command line, which is constructed using Typer.
At the moment typer does not support async functions, so we have to
use a wrapper to make it work. See utils.py for the AsyncTyper class.

The cli should be pretty self documenting, however see docs/ for
official documentation.
"""

import os

from gallagher import cc, __version__

from .utils import AsyncTyper

from .alarms import app as alarms_app
from .divisions import app as divisions_app
from .cardholders import app as cardholders_app
from .events import app as events_app

_help_text = """
gala is a command line interface for the gallagher security command centre.\n
It works by accessing and mutating objects on the command centre via it's\n
JSON API, directly or via the cloud gateway.\n
\n
It's intent it to make available scriptable endpoints to ease automation.
"""

# Load the API key for the package so all entry points
# can use it to query the service
api_key = os.environ.get("GACC_API_KEY")
cc.api_key = api_key

# Main Typer app use to create the CLI
app = AsyncTyper(
    help=_help_text,
)

# Load up all sub commands
app.add_typer(alarms_app, name="alarm")
app.add_typer(divisions_app, name="div")
app.add_typer(cardholders_app, name="ch")
app.add_typer(events_app, name="event")

if __name__ == "__main__":
    """In case you are invoking this via Python directly

    This is probably never actually used but it is here for completeness.
    You'd execute this by running `python -m gallagher.cli`
    """
    app()



from typing import Optional

from ..utils import (
    AppBaseModel,
)

from ..ref import (
    FeatureAccessGroupsRef,
    FeatureAccessZonesRef,
    FeatureAlarmZonesRef,
    FeatureAlarmsRef,
    FeatureCardTypesRef,
    FeatureCardholdersRef,
    FeatureCompetenciesRef,
    FeatureDayCategoriesRef,
    FeatureDivisionsRef,
    FeatureDoorsRef,
    FeatureElevatorsRef,
    FeatureEventsRef,
    FeatureFenceZonesRef,
    FeatureInputsRef,
    FeatureInterlockGroupsRef,
    FeatureItemsRef,
    FeatureLockerBanksRef,
    FeatureMacrosRef,
    FeatureOperatorGroupsRef,
    FeatureOutputsRef,
    FeaturePersonalDataFieldsRef,
    FeatureReceptionsRef,
    FeatureRolesRef,
    FeatureSchedulesRef,
    FeatureVisitsRef,
)


class FeaturesDetail(
    AppBaseModel,
):
    """A detailed list of features that are available on the server.

    All features are marked as Optional, which means that by default
    it's assumed that they are not available on the server. Upon discovery
    if a feature is enabled on the server then we receive a href which
    indicates to the client that the feature is available.

    If a feature is unavailable the API client will throw an exception.
    """

    access_groups: Optional[FeatureAccessGroupsRef] = FeatureAccessGroupsRef()
    access_zones: Optional[FeatureAccessZonesRef] = FeatureAccessZonesRef()
    alarm_zones: Optional[FeatureAlarmZonesRef] = FeatureAlarmZonesRef()
    alarms: Optional[FeatureAlarmsRef] = FeatureAlarmsRef()
    card_types: Optional[FeatureCardTypesRef] = FeatureCardTypesRef()
    cardholders: Optional[FeatureCardholdersRef] = FeatureCardholdersRef()
    competencies: Optional[FeatureCompetenciesRef] = FeatureCompetenciesRef()
    day_categories: Optional[FeatureDayCategoriesRef] = FeatureDayCategoriesRef()
    divisions: Optional[FeatureDivisionsRef] = FeatureDivisionsRef()
    doors: Optional[FeatureDoorsRef] = FeatureDoorsRef()
    elevators: Optional[FeatureElevatorsRef] = FeatureElevatorsRef()
    events: Optional[FeatureEventsRef] = FeatureEventsRef()
    fence_zones: Optional[FeatureFenceZonesRef] = FeatureFenceZonesRef()
    inputs: Optional[FeatureInputsRef] = FeatureInputsRef()
    interlock_groups: Optional[FeatureInterlockGroupsRef] = FeatureInterlockGroupsRef()
    items: Optional[FeatureItemsRef] = FeatureItemsRef()
    locker_banks: Optional[FeatureLockerBanksRef] = FeatureLockerBanksRef()
    macros: Optional[FeatureMacrosRef] = FeatureMacrosRef()
    operator_groups: Optional[FeatureOperatorGroupsRef] = FeatureOperatorGroupsRef()
    outputs: Optional[FeatureOutputsRef] = FeatureOutputsRef()
    personal_data_fields: Optional[FeaturePersonalDataFieldsRef] = (
        FeaturePersonalDataFieldsRef()
    )
    receptions: Optional[FeatureReceptionsRef] = FeatureReceptionsRef()
    roles: Optional[FeatureRolesRef] = FeatureRolesRef()
    schedules: Optional[FeatureSchedulesRef] = FeatureSchedulesRef()
    visits: Optional[FeatureVisitsRef] = FeatureVisitsRef()



from ..utils import (
    AppBaseResponseModel,
    HrefMixin,
)

from ..summary import (
    AlarmSummary,
)


class AlarmSummaryResponse(
    AppBaseResponseModel,
):
    """AlarmSummaryResponse represents a single alarm"""

    alarms: list[AlarmSummary]
    updates: HrefMixin

    @property
    def result_set(self) -> list[AlarmSummary]:
        """ Wrap summary response target property

        the sql interface will call this property and each summary
        response is expected to override this and return the appropriate
        target property
        """
        return self.alarms

    @property
    def cli_header(self):
        return ("id", "priority", "time", "type")

    def __rich_repr__(self):
        return (r.__rich_repr__() for r in self.alarms)

    def __str__(self):
        return f"{len(self.results)} cardholders"




""" Cardholder related 

"""

from ..core import Capabilities, APIEndpoint, EndpointConfig

from ...dto.detail import (
    CardholderDetail,
    PdfDetail,
)

from ...dto.response import (
    CardholderSummaryResponse,
    PdfResponse,
)


class Cardholder(APIEndpoint):
    """Cardholder endpoints allow you to search for and retrieve cardholder details.

    Cardholders are the users of the system and are the entities that are
    granted access to doors. Cardholders can be people, vehicles, or other
    entities that require access to the site.
    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.cardholders.cardholders,
            dto_list=CardholderSummaryResponse,
            dto_retrieve=CardholderDetail,
        )

    @classmethod
    async def search(cls, name: str, sort: str = "id", top: int = 100):
        pass


class PdfDefinition(APIEndpoint):
    """PDF Definitions provide a list of support PDF definitions for the instance.

    These can vary between using physical cards, mobile credentials, or
    biometrics. The card type is used to dynamically determine the types
    of credentials available on this particular instance.
    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.personal_data_fields\
                .personal_data_fields,
            dto_list=PdfResponse,
            dto_retrieve=PdfDetail,
        )

    @classmethod
    async def search(cls, name: str, sort: str = "id", top: int = 100):
        pass



__shillelagh__ = (
    Cardholder,
)


""" Reference classes

Our design outlines three types of DTO classes, Refs are ones that are
used to reference other objects, think of them as interlinks between
objects. They usually contain a href and some identifying information
such as a name or id.

This package was introduced in reference to this issue 
https://github.com/anomaly/gallagher/issues/21

which identified race conditions with circular imports. This is caused
mostly because of the nature of the data that the command centre exposes.
"""

from .access_group import AccessGroupRef

from .alarm import (
    AlarmRef,
    AlarmZoneRef,
)

from .card_type import (
    CardTypeRef,
)

from .cardholder import (
    CardholderRef,
    CardholderExtendedRef,
)

from .day_category import (
    DayCategoryRef,
)

from .discover import (
    FeatureAccessGroupsRef,
    FeatureAccessZonesRef,
    FeatureAlarmZonesRef,
    FeatureAlarmsRef,
    FeatureCardTypesRef,
    FeatureCardholdersRef,
    FeatureCompetenciesRef,
    FeatureDayCategoriesRef,
    FeatureDivisionsRef,
    FeatureDoorsRef,
    FeatureElevatorsRef,
    FeatureEventsRef,
    FeatureFenceZonesRef,
    FeatureInputsRef,
    FeatureInterlockGroupsRef,
    FeatureItemsRef,
    FeatureLockerBanksRef,
    FeatureMacrosRef,
    FeatureOperatorGroupsRef,
    FeatureOutputsRef,
    FeaturePersonalDataFieldsRef,
    FeatureReceptionsRef,
    FeatureRolesRef,
    FeatureSchedulesRef,
    FeatureVisitsRef,
)

from .division import (
    DivisionRef,
)

from .door import (
    DoorRef,
)

from .event import (
    EventGroupRef,
)

from .ext import (
    PlaceholderRef,
)

from .items import (
    ItemRef,
)

from .pdf import (
    PDFRef,
)

from .salto import (
    SaltoItemTypeRef,
    SaltoItemRef,
)

from .schedule import (
    ScheduleRef,
)


from .misc import (
    InstructionRef,
)

from .operator import (
    OperatorRef,
)

from .role import (
    RoleRef,
)

from .zone import (
    AccessZoneRef,
)



""" Asyncio patches for Typer

There's a thread open on the typer repo about this:
    https://github.com/tiangolo/typer/issues/88

Essentially, Typer doesn't support async functions. This is an issue
post migration to async, @csheppard points out that there's a work
around using a decorator on the click repository:
https://github.com/tiangolo/typer/issues/88#issuecomment-612687289

@gilcu2 posted a similar solution on the typer repo:
https://github.com/tiangolo/typer/issues/88#issuecomment-1732469681

this particular one uses asyncer to run the async function in a thread
we're going in with this with the hope that the official solution is
closer to this than a decorator per command.
"""

import inspect

from functools import (
    partial,
    wraps,
)

import asyncer
from typer import Typer


class AsyncTyper(Typer):
    @staticmethod
    def maybe_run_async(decorator, f):
        if inspect.iscoroutinefunction(f):

            @wraps(f)
            def runner(*args, **kwargs):
                return asyncer.runnify(f)(*args, **kwargs)

            decorator(runner)
        else:
            decorator(f)
        return f

    def callback(self, *args, **kwargs):
        decorator = super().callback(*args, **kwargs)
        return partial(self.maybe_run_async, decorator)

    def command(self, *args, **kwargs):
        decorator = super().command(*args, **kwargs)
        return partial(self.maybe_run_async, decorator)



from typing import Optional

from ..utils import (
    HrefMixin,
    IdentityMixin,
    AppBaseResponseModel,
    AppBaseResponseWithFollowModel,
)

from ..summary import (
    EventGroupSummary,
    EventSummary,
)


class EventSummaryResponse(
    AppBaseResponseWithFollowModel,
):

    events: list[EventSummary]


class EventTypeResponse(
    AppBaseResponseModel,
):
    """Event Type Response

    Event Type Responses return a set of eventGroups which in turn
    has identifiers, names and event types.

    """

    event_groups: list[EventGroupSummary]



from typing import Optional

from ..utils import (
    HrefMixin,
    AppBaseResponseWithFollowModel,
)

from ..detail import CardTypeDetail


class CardTypeResponse(
    AppBaseResponseWithFollowModel,
):
    """Card Types are cards mobile or physical that are supported at a site"""

    results: list[CardTypeDetail]



""" A day category links a calendar to a schedule. 

The calendar determines the days of the year that fall into a day category, 
and the schedule determines what happens at certain times on those days.
"""

from typing import Optional

from ..utils import AppBaseModel, HrefMixin


class DayCategorySummary(
    AppBaseModel,
    HrefMixin,
):
    """Represents a single entry from the response"""

    name: str
    description: Optional[str]
    notes: Optional[str]



from typing import Optional

from ..utils import (
    HrefMixin,
    AppBaseResponseWithFollowModel,
)

from ..summary import DivisionSummary

from ..detail import DivisionDetail


class DivisionSummaryResponse(
    AppBaseResponseWithFollowModel,
):
    """Division Summary

    This should return the summary and as per the documentation
    the detail should have more than what we have here

    TODO: check this is different

    """

    results: list[DivisionSummary]

    @property
    def cli_header(self):
        return (
            "id",
            "name",
            "server display name",
        )

    def __rich_repr__(self):
        return (r.__rich_repr__() for r in self.results)

    def __str__(self):
        return f"{len(self.results)} divisions"



from typing import Optional

from ..utils import (
    HrefMixin,
    AppBaseResponseWithFollowModel,
)

from ..ref import (
    DayCategoryRef,
)


class DayCategoryResponse(
    AppBaseResponseWithFollowModel,
):
    """The response has a list of results and a link to the next page"""

    results: list[DayCategoryRef]



""" Summary


"""

from .access_group import (
    AccessSummary,
    AccessGroupSummary,
    CardholderAccessGroupSummary,
)

from .alarm import (
    AlarmSourceSummary,
    AlarmSummary,
)

from .card_type import (
    CardExpiryTypeSummary,
    CardSummary,
)

from .cardholder import (
    CardholderSummary,
    CardholderCardSummary,
)

from .day_category import (
    DayCategorySummary,
)

from .division import DivisionSummary

from .door import (
    DoorSummary,
)

from .event import (
    EventSummary,
    EventGroupSummary,
    EventTypeSummary,
)

from .items import (
    ItemTypeSummary,
    ItemSummary,
)

from .salto import (
    SaltoAccessItemSummary,
)

from .schedule import (
    ScheduleSummary,
)

from .visitor import (
    VisitorManagementSummary,
    VisitorTypeSummary,
)

from .pdf import (
    PdfSummary,
)



""" Dashboard

Where you land up once the tui is started via the command line
this shows you some statistics, and health information.
"""

from textual.app import ComposeResult

from textual.containers import (
    Container,
    Grid,
)

from textual.widgets import (
    Digits,
    Placeholder,
)


class VisitTimelineWidget(Container):

    def compose(self) -> ComposeResult:
        """Visit timeline widget"""
        with Grid():
            yield Placeholder("Visit Timeline, changes")

class Dashboard(Container):

    CSS_PATH = "gallagher.tcss"

    def compose(self) -> ComposeResult:
        """Dashboard for the command centre"""
        with Grid():
            yield VisitTimelineWidget()